--[[
    Lua doesn't have type definitions, so no classes, etc. Documentation has to do.

    Since these are "Suits," to go with the addon name, the parts must be "Articles" (of clothing) and instead
    of "classes" or "types" we have "patterns," which fits better with the metatable/almost-prototype inheritance
    of Lua anyway.

    We have the Articles "namespace." This is never "instantiated." It contains...

    Methods:
        FromData    Takes a data bag table and sticks the Article table on as the metatable index, pasting the
                    methods onto the data. We could also do this using the GOF Flyweight pattern, and may change to
                    that later.
        FromType    Takes a type base (the table with TypeName and methods or just the type index) and returns an
                    inialized table that works as a functioning article of that type.

    Within the Articles "namespace" (rows of the Articles table), we have several tables that represent
    concrete kinds of the abstract article "type." These are then "instantiated" to represent an Article,
    one component of a Suit (think "article of clothing").

    An Article ("type" and "instance") has...

    Properties:
        Type    One of the ArticleType enumeration identifies which Article to use
        Data    Opaque data bag (table) which is specific to the type of article


    Static Properties: (Of course, there is no "static" in Lua. These are Article fields that don't need the Data field)
        TypeName    Human-readable name for the type of article: "Spec", "Chat", etc.

    Instance Methods
        Describe    Returns a string with some human-readable description of what the Article does in this
                    instance. This requires the Data to be present and valid, probably.
        DescribeAML Returns an Apollo Markup Language version of the Describe text
        IsEnabled   Returns true if the Article should be shown for adding to suits, should be configurable, and should Apply
        Apply       Actually makes the changes that the Article represents: swap specs, say something in chat,
                    change settings, etc.
        Init        Builds valid, minimal data values for a new article instance of this type
        MakeEditor  Arguments: hwndParent
                    Loads and initializes an editor form for the Article

--]]
local Patterns = {}

--region Constants
local TYPE = {TABLE=type({}),NUMBER=type(1),STRING=type("a"),BOOL=type(true),FUNCTION=type(function () end)}

local EDITOR_FONT = "DefaultButton"
--endregion

--region Enums, Display sequences, and Names
--[[=============================================================================================]]
-- The types of Articles in a Suit. Each has a table of methods and expects some members with data
-- (basically, it's like a class).
Patterns.eArticleType =
{
    Costume     = 1,
    Spec        = 2,
    Say         = 3,
    Nameplates  = 4,
    Telegraphs  = 5,
--  Mount       = 6,
    Title       = 7,
    GuildHolo   = 8,
    PvP         = 9,
    CommandLine = 10,
    Stance      = 11,
    Dash        = 12,
    Affiliation = 13,
    Macro       = 14,
    Targeting   = 15,
}

Patterns.ArticleTypeNames =
{
    [Patterns.eArticleType.Spec]        = "Action Sets",
    [Patterns.eArticleType.Dash]        = "Dash On/Off",
    [Patterns.eArticleType.Costume]     = "Costumes",
    [Patterns.eArticleType.Say]         = "Broadcast Text",
    [Patterns.eArticleType.Stance]      = "Stance",
    [Patterns.eArticleType.CommandLine] = "Slash Command",
    [Patterns.eArticleType.Telegraphs]  = "Telegraph Settings",
    [Patterns.eArticleType.Title]       = "Titles",
    [Patterns.eArticleType.GuildHolo]   = "Holomark Controls",
    [Patterns.eArticleType.PvP]         = "PvP Flag On/Off",
    [Patterns.eArticleType.Affiliation] = "Your Nameplate Affiliation",
    [Patterns.eArticleType.Nameplates]  = "Nameplates Display",
    [Patterns.eArticleType.Macro]       = "Macros",
    [Patterns.eArticleType.Targeting]   = "Targeting Settings",
}

Patterns.ArticleTypeIcons =
{
    [Patterns.eArticleType.Spec]        = "HUD_BottomBar:spr_HUD_MenuIcons_Abilities",
    [Patterns.eArticleType.Dash]        = "CRB_ActionBarFrameSprites:sprResourceBar_Sprint_RunIconBlue",
    [Patterns.eArticleType.Costume]     = "BK3:sprHolo_Icon_Costume",
    [Patterns.eArticleType.Say]         = "IconSprites:Icon_MapNode_Map_Chat",
    [Patterns.eArticleType.Stance]      = "IconSprites:Icon_SkillStalker_Stance_Defensive",
    [Patterns.eArticleType.CommandLine] = "IconSprites:Icon_TradeskillAdditiveArchitect_Keypad_00",
    [Patterns.eArticleType.Telegraphs]  = "CRB_ChallengeSprites:sprCh_SpinAnimation_1",
    [Patterns.eArticleType.Title]       = "IconSprites:Icon_Windows_UI_CRB_Marker_Mask",
    [Patterns.eArticleType.GuildHolo]   = "IconSprites:Icon_Modifier_resist_tech_001",
    [Patterns.eArticleType.PvP]         = "HUD_TargetFrame:spr_TargetFrame_ClassIcon_Elite",
    [Patterns.eArticleType.Affiliation] = "IconSprites:Icon_Windows_UI_CRB_Marker_Octopus",
    [Patterns.eArticleType.Nameplates]  = "CRB_Nameplates:sprNP_BaseSelectedRed",
    [Patterns.eArticleType.Macro]       = "IconSprites:Icon_Windows32_UI_CRB_InterfaceMenu_Macro",
    [Patterns.eArticleType.Targeting]   = "Crafting_CoordSprites:sprCoord_AdditiveTargetRed",
}

Patterns.ArticleDisplayOrder =
{
    Patterns.eArticleType.Spec,
    Patterns.eArticleType.Say,
    Patterns.eArticleType.Costume,
    Patterns.eArticleType.Dash,
    Patterns.eArticleType.GuildHolo,
    Patterns.eArticleType.Macro,
    Patterns.eArticleType.Nameplates,
    Patterns.eArticleType.PvP,
    Patterns.eArticleType.CommandLine,
    Patterns.eArticleType.Stance,
    Patterns.eArticleType.Targeting,
    Patterns.eArticleType.Telegraphs,
    Patterns.eArticleType.Title,
    Patterns.eArticleType.Affiliation,
}

-- All the channels defined in ChatLog.Lua. Note that some are not useful to us (or may not actually exist).
-- It turns out that these are not channels. They are channel types. There is no good list of channels, and some types may apply to
-- more than one channel.
local ChatChannels =
{
  [ChatSystemLib.ChatChannel_Say            ] = {DisplayName = "Say",               Channel=ChatSystemLib.ChatChannel_Say            , Private = false},   --4,
  [ChatSystemLib.ChatChannel_Party          ] = {DisplayName = "Party",             Channel=ChatSystemLib.ChatChannel_Party          , Private = false},   --7,
  [ChatSystemLib.ChatChannel_Emote          ] = {DisplayName = "Emote",             Channel=ChatSystemLib.ChatChannel_Emote          , Private = false},   --8,
  [ChatSystemLib.ChatChannel_Guild          ] = {DisplayName = "Guild",             Channel=ChatSystemLib.ChatChannel_Guild          , Private = false},   --15,
  [ChatSystemLib.ChatChannel_GuildOfficer   ] = {DisplayName = "Guild Officer",     Channel=ChatSystemLib.ChatChannel_GuildOfficer   , Private = false},   --16,
  [ChatSystemLib.ChatChannel_Instance       ] = {DisplayName = "Instance",          Channel=ChatSystemLib.ChatChannel_Instance       , Private = false},   --32,
  [ChatSystemLib.ChatChannel_Yell           ] = {DisplayName = "Yell",              Channel=ChatSystemLib.ChatChannel_Yell           , Private = false},   --5,
  [ChatSystemLib.ChatChannel_Zone           ] = {DisplayName = "Zone",              Channel=ChatSystemLib.ChatChannel_Zone           , Private = false},   --9,
  [ChatSystemLib.ChatChannel_ZonePvP        ] = {DisplayName = "Zone PvP",          Channel=ChatSystemLib.ChatChannel_ZonePvP        , Private = false},   --10,
  [ChatSystemLib.ChatChannel_WarParty       ] = {DisplayName = "War Party",         Channel=ChatSystemLib.ChatChannel_WarParty       , Private = false},   --29,
  [ChatSystemLib.ChatChannel_WarPartyOfficer] = {DisplayName = "War Party Officer", Channel=ChatSystemLib.ChatChannel_WarPartyOfficer, Private = false},   --30,
  [ChatSystemLib.ChatChannel_Advice         ] = {DisplayName = "Advice",            Channel=ChatSystemLib.ChatChannel_Advice         , Private = false},   --33,
  [ChatSystemLib.ChatChannel_Trade          ] = {DisplayName = "Trade",             Channel=ChatSystemLib.ChatChannel_Trade          , Private = false},   --11,
  [ChatSystemLib.ChatChannel_Command        ] = {DisplayName = "Command",           Channel=ChatSystemLib.ChatChannel_Command        , Private = true},   --1,
  [ChatSystemLib.ChatChannel_Realm          ] = {DisplayName = "Realm",             Channel=ChatSystemLib.ChatChannel_Realm          , Private = true,},   --25,
  [ChatSystemLib.ChatChannel_Datachron      ] = {DisplayName = "Datachron",         Channel=ChatSystemLib.ChatChannel_Datachron      , Private = true,},   --23,
  [ChatSystemLib.ChatChannel_Debug          ] = {DisplayName = "Debug",             Channel=ChatSystemLib.ChatChannel_Debug          , Private = true,},   --3,
--[[
  [ChatSystemLib.ChatChannel_AccountWhisper ] = {DisplayName = "Account Whisper",   Channel=ChatSystemLib.ChatChannel_AccountWhisper , Private = },   --34,
  [ChatSystemLib.ChatChannel_ActionEmote    ] = {DisplayName = "Action Emote",      Channel=ChatSystemLib.ChatChannel_ActionEmote    , Private = },   --28,
  [ChatSystemLib.ChatChannel_AnimatedEmote  ] = {DisplayName = "Animated Emote",    Channel=ChatSystemLib.ChatChannel_AnimatedEmote  , Private = },   --27,
  [ChatSystemLib.ChatChannel_Class          ] = {DisplayName = "Class",             Channel=ChatSystemLib.ChatChannel_Class          , Private = },   --13,
  [ChatSystemLib.ChatChannel_Combat         ] = {DisplayName = "Combat",            Channel=ChatSystemLib.ChatChannel_Combat         , Private = },   --24,
  [ChatSystemLib.ChatChannel_Custom         ] = {DisplayName = "Custom",            Channel=ChatSystemLib.ChatChannel_Custom         , Private = },   --18,
  [ChatSystemLib.ChatChannel_GuildRecruit   ] = {DisplayName = "Guild Recruit",     Channel=ChatSystemLib.ChatChannel_GuildRecruit   , Private = },   --14,
  [ChatSystemLib.ChatChannel_LFG            ] = {DisplayName = "LFG",               Channel=ChatSystemLib.ChatChannel_LFG            , Private = },   --12,
  [ChatSystemLib.ChatChannel_Loot           ] = {DisplayName = "Loot",              Channel=ChatSystemLib.ChatChannel_Loot           , Private = },   --26,
  [ChatSystemLib.ChatChannel_NPCSay         ] = {DisplayName = "NPC Say",           Channel=ChatSystemLib.ChatChannel_NPCSay         , Private = },   --20,
  [ChatSystemLib.ChatChannel_NPCWhisper     ] = {DisplayName = "NPC Whisper",       Channel=ChatSystemLib.ChatChannel_NPCWhisper     , Private = },   --22,
  [ChatSystemLib.ChatChannel_NPCYell        ] = {DisplayName = "NPC Yell",          Channel=ChatSystemLib.ChatChannel_NPCYell        , Private = },   --21,
  [ChatSystemLib.ChatChannel_PlayerPath     ] = {DisplayName = "Player Path",       Channel=ChatSystemLib.ChatChannel_PlayerPath     , Private = },   --31,
  [ChatSystemLib.ChatChannel_Society        ] = {DisplayName = "Society",           Channel=ChatSystemLib.ChatChannel_Society        , Private = },   --17,
  [ChatSystemLib.ChatChannel_Support        ] = {DisplayName = "Support",           Channel=ChatSystemLib.ChatChannel_Support        , Private = },   --19,
  [ChatSystemLib.ChatChannel_System         ] = {DisplayName = "System",            Channel=ChatSystemLib.ChatChannel_System         , Private = },   --2,
  [ChatSystemLib.ChatChannel_Whisper        ] = {DisplayName = "Whisper",           Channel=ChatSystemLib.ChatChannel_Whisper        , Private = },   --6,
--]]
}
Patterns.ChatChannelDisplayOrder =
{
    ChatSystemLib.ChatChannel_Say,
    ChatSystemLib.ChatChannel_Party,
    ChatSystemLib.ChatChannel_Instance,
    ChatSystemLib.ChatChannel_Emote,
    ChatSystemLib.ChatChannel_Guild,
    ChatSystemLib.ChatChannel_GuildOfficer,
    ChatSystemLib.ChatChannel_Yell,
    ChatSystemLib.ChatChannel_Zone,
    ChatSystemLib.ChatChannel_ZonePvP,
    ChatSystemLib.ChatChannel_WarParty,
    ChatSystemLib.ChatChannel_WarPartyOfficer,
    ChatSystemLib.ChatChannel_Advice,
    ChatSystemLib.ChatChannel_Trade,
    ChatSystemLib.ChatChannel_Command,
    ChatSystemLib.ChatChannel_Realm,
    ChatSystemLib.ChatChannel_Datachron,
    ChatSystemLib.ChatChannel_Debug,
}
--endregion

--region Helper Functions
-- Returns the lowest point of any child of wnd, measured from the top
local function GetWindowFooter(wnd)
    local lowest = 0
    for k,v in pairs(wnd:GetChildren()) do
        left, top, right, bottom = v:GetRect()
        lowest = math.max(lowest, bottom)
    end
    return lowest
end

--[[
local function ResizeWindowHeight(wnd, margin)
    margin = margin or 0
    local l, t, r, b = wnd:GetAnchorOffsets()
    local lA, tA, rA, bA = wnd:GetAnchorPoints()

    if bA == 0 then b = GetWindowFooter(wnd) + margin end
    --if bA == 1 then b =
    wnd:SetAnchorOffsets(l, t, r, b)
end
--]]
--endregion Helper Functions

--region ArticleHelpers
--[[=============================================================================================]]
local ArticleHelpers = {}

-- Helper to make a select editor, set up the label, fill the combo with some options, and select something
-- The options list is {Index = {Text, Data} ...}
function ArticleHelpers:MakeSelectEditor(wndParent, txtLabel, tOptions, DefaultData, tEventHandler)
    local CL = Apollo.GetPackage("Xowl:LibCombo-1.0").tPackage

    local wndEditor = Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorSelect", wndParent, tEventHandler or self)
    wndEditor:FindChild("ArticleEditorSelectLabel"):SetText(txtLabel or "")

    local wndCombo = wndEditor:FindChild("ArticleEditorSelectCombo")

    -- Resize the label so the text shows properly, then move the left edge of the combo next to it
    local nWidth = Apollo.GetTextWidth(EDITOR_FONT, txtLabel)
    local leftLabel, topLabel, rightLabel, bottomLabel = wndEditor:FindChild("ArticleEditorSelectLabel"):GetAnchorOffsets()
    wndEditor:FindChild("ArticleEditorSelectLabel"):SetAnchorOffsets(leftLabel, topLabel, leftLabel + nWidth + 5, bottomLabel)

    local leftCombo, topCombo, rightCombo, bottomCombo = wndCombo:GetAnchorOffsets()
    wndCombo:SetAnchorOffsets(leftLabel + nWidth + 12, topCombo, rightCombo, bottomCombo)

    -- Create the new combo control
    local deftext = ""
    for _,def in ipairs(tOptions) do
        if def.Data == DefaultData then deftext = def.Text; break; end
    end

    CL:Create(wndCombo, tOptions, deftext, nil, tEventHandler)

end

-- A lot of the editors need a 1-based numeric selector. This helpers makes it easy.
function ArticleHelpers:MakeIndexSelectEditor(wndParent, txtLabel, nMaxIndex, DefaultIndex, tEventHandler)
    local tIndexes = {}
    for i = 1,nMaxIndex do
        table.insert(tIndexes, {Text = tostring(i), Data = i})
    end
    return self:MakeSelectEditor(wndParent, txtLabel, tIndexes, DefaultIndex, tEventHandler)
end
--endregion

--region Article template (abstract type definition)
--[[=============================================================================================]]
local Articles = {}
--[[
-- Create(type) returns an article type (with methods attached) that has the base most-neutral data possible
function Articles:Create(TypeId)
    local ret = self:FromData({ArticleType = TypeId,})
    ret:Init()
    return ret
end
--]]

-- Take an article type table or type number and make a new data table that a) has the type as a metatable index and b) is initialized
-- Note that this does't actually need the self parameter. Left it on for consistency. There is enough confusion between . and : accessors as-is.
function Articles:FromType(Type)
    local ret = {}
    if type(Type) == TYPE.TABLE then
        setmetatable(ret, {__index == Type}) --This doesn't seem to work right with tables I expect to work, but it's easy to avoid using
    elseif type(Type) == TYPE.NUMBER then
        setmetatable(ret, {__index = Articles[Type]})
    elseif type(Type) == TYPE.STRING then
        local n = tonumber(Type)
        if n ~= nil then
            setmetatable(ret, {__index = Articles[n]})
        end
    end
    ret:Init()
    return ret
end

-- Call Articles:FromData(ArticleData) to load the data into a fresh article. ArticleData is assumed to be a table with
-- ArticleType, ArticleVer, and any data the article needs
function Articles:FromData(ArticleData)
    local ret = self:FromType(ArticleData.ArticleType)
    ret:InitFromData(ArticleData)
    return ret
end


--[[=============================================================================================]]
-- Specific article patterns
-- First, a generic one we can copy and modify (prototype-based inheritance style)
--[[
    A basic template to cut-and-paste
local ___Pattern = {}
___Pattern = NewArticlePattern(Patterns.eArticleType.____, "____")
function :Init()                end;
function :InitFromData(data)    end;
function :ToData()              return {ArticleType = self.ArticleType, ArticleVersion = 1}     end;
function :Describe()            return ""; end;
function :Apply()               end;
function :MakeEditor(wndParent) end;
function :IsEnabled()           return true; end;
Articles[Patterns.eArticleType.___] = ___Pattern
--]]
local BaseCallError = function() Print("SuitUp Error: Base article pattern called somehow.") end
local ArticlePatternTemplate =
{
    TypeName     = nil,
    Init         = BaseCallError,
    InitFromData = function(self) return; end,
    ToData       = function(self) return {ArticleType = self.ArticleType}; end,
    Apply        = BaseCallError,
    Describe     = BaseCallError,
    DescribeAML  = function(self) return "<P>"..self:Describe().."</P>"; end, -- We have a default for this one
    MakeEditor   = BaseCallError,
    IsEnabled    = function(self) return true; end,
}

---------------------------------------------------------------------------------------------------
local function NewArticlePattern(ArticleType)
    local ret = {}
    setmetatable(ret, {__index = ArticlePatternTemplate})
    ret.ArticleType = ArticleType
    ret.TypeName = Patterns.ArticleTypeNames[ArticleType]
    if ret.TypeName == nil or ret.TypeName == "" then Print("SuitUp error: No description for action type #"..tostring(ArticleType)); end
    return ret
end


--[[=============================================================================================--
    Now, we'd *like* to do something more elegant like this:
        Articles[id] = NewArticlePattern(Name)
        function Articles[id]:Foo() ... end

    But, the BabeLua extension for Visual Studio doesn't know how to parse that as a type declaration,
    nor would I really expect it to. So it makes the outlining unhappy and makes it harder to navigate
    the file in the editor.

    Instead we're using this slightly-annoying idiom:
        local temp = {}
        temp = NewArticlePattern(Name)
        function temp:Foo ... end
        Articles[id] = temp
--]]

--endregion


--region Dash
--[[=============================================================================================]]
--[[
    Data:
        EnableDash  Boolean     Whether to set Dash on or off. Defaults to on.
--]]
local DashPattern = {}
DashPattern = NewArticlePattern(Patterns.eArticleType.Dash)

---------------------------------------------------------------------------------------------------
function DashPattern:Init()
    self.EnableDash = true
end

---------------------------------------------------------------------------------------------------
function DashPattern:InitFromData(data)
    self.EnableDash = data.EnableDash
end

---------------------------------------------------------------------------------------------------
function DashPattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, EnableDash = self.EnableDash}
end

---------------------------------------------------------------------------------------------------
function DashPattern:Apply()
    Apollo.SetConsoleVariable("player.doubleTapToDash", self.EnableDash)
end

---------------------------------------------------------------------------------------------------
function DashPattern:Describe()
    local rets =
    {
     [true]  = "Enable double-tap move keys to dash",
     [false] = "Disable double-tap move keys to dash"
    }

    return rets[self.EnableDash]
end

---------------------------------------------------------------------------------------------------
function DashPattern:MakeEditor(wndParent)
    ArticleHelpers:MakeSelectEditor(wndParent, "Enable or disable double-tap to dash", {{Text="Enable", Data=true}, {Text="Disable", Data=false}}, self.EnableDash, self)
end

---------------------------------------------------------------------------------------------------
function DashPattern:OnXCSelect(control_data, row_data)
    self.EnableDash = row_data.Data
end

Articles[Patterns.eArticleType.Dash] = DashPattern
--endregion

--region Costume
--[[=============================================================================================]]
--[[
    Data:
        CostumeId   Number  Which costume to change to
--]]
local CostumePattern = {}
CostumePattern = NewArticlePattern(Patterns.eArticleType.Costume)

---------------------------------------------------------------------------------------------------
function CostumePattern:Init()
    self.CostumeId = 1
end

---------------------------------------------------------------------------------------------------
function CostumePattern:InitFromData(data)
    if data.ArticleVersion ~= 1 then Print("SuitUp: Invalid save data"); end;
    self.CostumeId = data.CostumeId
end

---------------------------------------------------------------------------------------------------
function CostumePattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, CostumeId = self.CostumeId}
end

---------------------------------------------------------------------------------------------------
function CostumePattern:Apply()
    local nCostumeId = tonumber(self.CostumeId)
    GameLib.SetCostumeIndex(nCostumeId)
end

---------------------------------------------------------------------------------------------------
function CostumePattern:Describe()
    return "Wear costume #"..tostring(self.CostumeId)
end

---------------------------------------------------------------------------------------------------
function CostumePattern:MakeEditor(wndParent)
    local tOptions = {}
    for i = 1,GameLib.GetCostumeCount() do
        tOptions[i] = {Text = "Costume "..tostring(i), Data = i}
    end
    table.insert(tOptions, 1, {Text = "(No Costume)", Data = 0})
    ArticleHelpers:MakeSelectEditor(wndParent, "Costume Number", tOptions, self.CostumeId, self)
end

---------------------------------------------------------------------------------------------------
function CostumePattern:OnXCSelect(control_data, row_data)
    self.CostumeId = row_data.Data
end

---------------------------------------------------------------------------------------------------
Articles[Patterns.eArticleType.Costume] = CostumePattern
--endregion

--region Spec
--[[=============================================================================================]]
--[[
    Data:
        SpecId   Number  Which LAS spec to change to
--]]
local SpecPattern = {}
SpecPattern = NewArticlePattern(Patterns.eArticleType.Spec)

---------------------------------------------------------------------------------------------------
function SpecPattern:Init()
    self.SpecId = 1
end

---------------------------------------------------------------------------------------------------
function SpecPattern:InitFromData(data)
    if data.ArticleVersion ~= 1 then Print("SuitUp: Invalid save data"); end;
    self.SpecId = data.SpecId
end

---------------------------------------------------------------------------------------------------
function SpecPattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, SpecId = self.SpecId}
end

---------------------------------------------------------------------------------------------------
function SpecPattern:Describe()
    return "Change to Action Set #"..tonumber(self.SpecId)
end

---------------------------------------------------------------------------------------------------
function SpecPattern:Apply()
    local nSpec = tonumber(self.SpecId)
    if nSpec ~= AbilityBook.GetCurrentSpec and GameLib.GetPlayerUnit():IsInCombat() == true then
        -- Let the user know we couldn't change specs, but try it anyway and continue trying the rest of the actions
        --TODO: Inform user
    end
    AbilityBook.SetCurrentSpec(nSpec)
end

---------------------------------------------------------------------------------------------------
function SpecPattern:MakeEditor(wndParent)
    ArticleHelpers:MakeIndexSelectEditor(wndParent, "Action Set Number", AbilityBook.GetNumUnlockedSpecs(), self.SpecId, self)
end

---------------------------------------------------------------------------------------------------
function SpecPattern:OnXCSelect(control_data, row_data)
    self.SpecId = row_data.Data
end

---------------------------------------------------------------------------------------------------
Articles[Patterns.eArticleType.Spec] = SpecPattern
--endregion Spec

--region Say/Chat
--[[=============================================================================================]]
--[[
    Data:
        Message String      The text to send
        Channel Table       {DisplayName, Channel}
        Percent Number      Number between 0 and 1 representing the chance to say the Message
--]]
local SayPattern = {}
SayPattern = NewArticlePattern(Patterns.eArticleType.Say)

---------------------------------------------------------------------------------------------------
function SayPattern:Init()
    self.Message = ""
    self.Channel = ChatChannels[ChatSystemLib.ChatChannel_Say].Channel
    self.Percent = 1
end

---------------------------------------------------------------------------------------------------
function SayPattern:InitFromData(data)
    if data.ArticleVersion ~= 1 then Print("SuitUp: Invalid save data"); end;
    self.Message = data.Message
    self.Channel = data.Channel
    self.Percent = data.Percent
end

---------------------------------------------------------------------------------------------------
function SayPattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, Message = self.Message, Channel = self.Channel, Percent = self.Percent}
end

---------------------------------------------------------------------------------------------------
function SayPattern:Describe()
    if self.Percent == nil then self.Percent = 1; end
    if self.Percent >= 1 then return "["..ChatChannels[self.Channel].DisplayName.."] "..self.Message; end
    if self.Percent > 0  then return "["..ChatChannels[self.Channel].DisplayName..", "..tostring(math.floor(self.Percent * 100)).."%] "..self.Message; end
    return "(disabled) ["..ChatChannels[self.Channel].DisplayName.."] "..self.Message
end

---------------------------------------------------------------------------------------------------
function SayPattern:Apply()
    if self.Percent == nil then self.Percent = 1; end
    local pct = tonumber(self.Percent)
    if pct and pct >= 0 and pct < 1 and math.random() > self.Percent then return; end -- Failed the random chance
    if Apollo.GetAddon("SuitUp").tOptions.DisableChat and not ChatChannels[self.Channel].Private then return; end -- Squelched

    local channel = self.Channel
    
    -- If they are using chat fallback (from Instance -> Party -> Say as needed), find the appropriate channel
    if Apollo.GetAddon("SuitUp").tOptions.UseChatFallback then 
        -- Fallback from Instance to Party
        if not GroupLib.InInstance() and channel == ChatSystemLib.ChatChannel_Instance then
            channel = ChatSystemLib.ChatChannel_Party
        end

        -- Fallback from Party to Say
        if not GroupLib.InGroup() and channel == ChatSystemLib.ChatChannel_Party then
            channel = ChatSystemLib.ChatChannel_Say
        end
    end
    local cid = nil
    for i,v in ipairs(ChatSystemLib.GetChannels()) do
        if v:GetType() == channel then
            cid = i
            break;
        end
    end
    if cid == nil then Print("SuitUp: Error sending to channel"); return; end
    
    -- Private channels can't use the Send command. We just Post on them.
    if ChatChannels[channel].Private then
        ChatSystemLib.PostOnChannel(channel, self.Message)
        return
    end

    ChatSystemLib.GetChannels()[cid]:Send(self.Message)
end

---------------------------------------------------------------------------------------------------
function SayPattern:MakeEditor(wndParent)
    self.wndEditor = Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorChat", wndParent, self)

    local wndCombo = self.wndEditor:FindChild("ChannelCombo")

    local tOptions = {}
    for i,chid in ipairs(Patterns.ChatChannelDisplayOrder) do
        local chan = ChatChannels[chid]
        local dn = chan.DisplayName
        if chan.Private then dn = dn .. " (private channel)"; end
        table.insert(tOptions, {Text = dn, Data = chid})
    end

    local CL = Apollo.GetPackage("Xowl:LibCombo-1.0").tPackage
    CL:Create(wndCombo, tOptions, ChatChannels[self.Channel].DisplayName, nil, self)

    self.wndEditor:FindChild("MessageEdit"):SetText(self.Message)

    self:SetPercentControls()
end

---------------------------------------------------------------------------------------------------
function SayPattern:OnXCSelect(control_data, row_data)
    self.Channel = row_data.Data
end

---------------------------------------------------------------------------------------------------
function SayPattern:OnMessageChanged(wndHandler, wndControl, strText)
    self.Message = strText
end

---------------------------------------------------------------------------------------------------
function SayPattern:OnPercentEditChanged(wndHandler, wndControl, strText)
    local n = tonumber(strText)
    if n ~= nil then
        n = math.floor(n)
        if n >=0 and n <= 100 then
            -- All is good, we have a number between 0 and 100
            self.Percent = n / 100
        end
    end
    self:SetPercentControls()
end

---------------------------------------------------------------------------------------------------
function SayPattern:OnPercentSliderChanging(wndHandler, wndControl, fNewValue, fOldValue, bOkToChange)
    self.Percent = math.floor(fNewValue)/100
    self:SetPercentControls()
end

---------------------------------------------------------------------------------------------------
function SayPattern:SetPercentControls()
    self.wndEditor:FindChild("PercentEdit"):SetText(tostring(math.floor(self.Percent * 100)))
    self.wndEditor:FindChild("PercentEdit"):SetSel(string.len(tostring(math.floor(self.Percent * 100))))
    self.wndEditor:FindChild("PercentSlider"):SetValue(math.floor(self.Percent * 100))
end

---------------------------------------------------------------------------------------------------
Articles[Patterns.eArticleType.Say] = SayPattern
--endregion Say/Chat

--region Stance
--[[=============================================================================================]]
--[[
    Data:
        StanceId    Number      Stance number to set, as shown on the "Set Stance #" keybinding. Note that the index
                                into the stances table is different: tableindex = (StanceId * 2 - 1)
--]]
local StancePattern = {}
StancePattern = NewArticlePattern(Patterns.eArticleType.Stance)

---------------------------------------------------------------------------------------------------
function StancePattern:Init()
    self.StanceId = 1 -- Default
end

---------------------------------------------------------------------------------------------------
function StancePattern:InitFromData(data)
    self.StanceId = data.StanceId
end

---------------------------------------------------------------------------------------------------
function StancePattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, StanceId = self.StanceId}
end

---------------------------------------------------------------------------------------------------
function StancePattern:Describe()
    return "Activate '"..GameLib.GetClassInnateAbilitySpells().tSpells[self.StanceId * 2 - 1]:GetName().."'"
end

---------------------------------------------------------------------------------------------------
function StancePattern:IsEnabled()
    -- If they can queue as a tank, they have Stances to set. At least as of API 9. 
    -- If the healing classes ever get Stances, this should change.

    for i,v in ipairs(MatchingGame.GetEligibleRoles()) do
        if v == MatchingGame.Roles.Tank then return true; end
    end

    return false;
end

---------------------------------------------------------------------------------------------------
function StancePattern:Apply()
    GameLib.SetCurrentClassInnateAbilityIndex(self.StanceId)
end

---------------------------------------------------------------------------------------------------
function StancePattern:MakeEditor(wndParent)
    local Stances = {}

    local StanceCount = 0
    for idx, spellObject in pairs(GameLib.GetClassInnateAbilitySpells().tSpells) do
        if idx % 2 == 1 then
            StanceCount = StanceCount + 1
            Stances[StanceCount] = {Text = spellObject:GetName(), Data = StanceCount}
        end
    end
    ArticleHelpers:MakeSelectEditor(wndParent, "Activate Stance", Stances, self.StanceId, self)
end

---------------------------------------------------------------------------------------------------
function StancePattern:OnXCSelect(control_data, row_data)
    self.StanceId = row_data.Data
end

Articles[Patterns.eArticleType.Stance] = StancePattern
--endregion Stance

--region Command
--[[=============================================================================================]]
--[[
    Data:
        Command     String      The command line to execute
--]]
local CommandLinePattern = {}
CommandLinePattern = NewArticlePattern(Patterns.eArticleType.CommandLine)

function CommandLinePattern:Init()
    self.Command = nil
end

function CommandLinePattern:InitFromData(data)
    self.Command = data.Command
end

function CommandLinePattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, Command = self.Command}
end

function CommandLinePattern:Describe()
    if self.Command == nil then return "Execute command line (no command defined)"; end
    return "Execute command line: "..tostring(self.Command)
end

function CommandLinePattern:Apply()
    if type(self.Command) == TYPE.STRING then 
        local cmd = self.Command
        if "/" ~= string.sub(cmd, 1, 1) then
            cmd = "/"..cmd
        end
        ChatSystemLib.Command(cmd); 
    end
end

function CommandLinePattern:MakeEditor(wndParent)
    self.wndEditor =  Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorCommand", wndParent, self)
    if self.Command ~= nil then self.wndEditor:FindChild("CommandEdit"):SetText(self.Command); end
end

function CommandLinePattern:OnCommandChanged(wndHandler, wndControl, strText)
    self.Command = strText
end

Articles[Patterns.eArticleType.CommandLine] = CommandLinePattern
--endregion Command

--region Telegraphs
--[[=============================================================================================]]
--[[
    Data:
        ShowSelf            Boolean
        ShowEnemy           Boolean
        ShowEnemyNPCBene    Boolean
        ShowEnemyNPCHarm    Boolean
        ShowEnemyPCBene     Boolean
        ShowEnemyPCHarm     Boolean
        ShowAlly            Boolean
        ShowAllyPartyOnly   Boolean
        ShowAllyNPCBene     Boolean
        ShowAllyNPCHarm     Boolean
        ShowAllyPCBene      Boolean
        ShowAllyPCHarm      Boolean
--]]
local TelegraphsPattern = {}
TelegraphsPattern = NewArticlePattern(Patterns.eArticleType.Telegraphs)

-- Some helpful constants so we can make helpful helpers that use them helpfully
TelegraphsPattern.TelegraphOptions =
{
    SelfTelegraphs  = 1,
    EnemyTelegraphs = 2,
    EnemyNPCBene    = 3,
    EnemyNPCHarm    = 4,
    EnemyPCBene     = 5,
    EnemyPCHarm     = 6,
    AllyTelegraphs  = 7,
    AllyPartyOnly   = 8,
    AllyNPCBene     = 9,
    AllyNPCHarm     = 10,
    AllyPCBene      = 11,
    AllyPCHarm      = 12,
}

TelegraphsPattern.ControlMapping =
{
    [TelegraphsPattern.TelegraphOptions.SelfTelegraphs ]  = {ConsoleVariable = "spell.selfTelegraphDisplay",                    ControlName = "CheckTelegraphsSelf"},
    [TelegraphsPattern.TelegraphOptions.EnemyTelegraphs]  = {ConsoleVariable = "spell.enemyTelegraphDisplay",                   ControlName = "CheckTelegraphsEnemy"},
    [TelegraphsPattern.TelegraphOptions.EnemyNPCBene   ]  = {ConsoleVariable = "spell.enemyNPCBeneficialTelegraphDisplay",      ControlName = "CheckTelegraphsEnemyNPCBene"},
    [TelegraphsPattern.TelegraphOptions.EnemyNPCHarm   ]  = {ConsoleVariable = "spell.enemyNPCDetrimentalTelegraphDisplay",     ControlName = "CheckTelegraphsEnemyNPCHarm"},
    [TelegraphsPattern.TelegraphOptions.EnemyPCBene    ]  = {ConsoleVariable = "spell.enemyPlayerBeneficialTelegraphDisplay",   ControlName = "CheckTelegraphsEnemyPCBene"},
    [TelegraphsPattern.TelegraphOptions.EnemyPCHarm    ]  = {ConsoleVariable = "spell.enemyPlayerDetrimentalTelegraphDisplay",  ControlName = "CheckTelegraphsEnemyPCHarm"},
    [TelegraphsPattern.TelegraphOptions.AllyTelegraphs ]  = {ConsoleVariable = "spell.allyTelegraphDisplay",                    ControlName = "CheckTelegraphsAlly"},
    [TelegraphsPattern.TelegraphOptions.AllyPartyOnly  ]  = {ConsoleVariable = "spell.partyMemberAllyTelegraphDisplay",         ControlName = "CheckTelegraphsAllyPartyOnly"},
    [TelegraphsPattern.TelegraphOptions.AllyNPCBene    ]  = {ConsoleVariable = "spell.allyNPCBeneficialTelegraphDisplay",       ControlName = "CheckTelegraphsAllyNPCBene"},
    [TelegraphsPattern.TelegraphOptions.AllyNPCHarm    ]  = {ConsoleVariable = "spell.allyNPCDetrimentalTelegraphDisplay",      ControlName = "CheckTelegraphsAllyNPCHarm"},
    [TelegraphsPattern.TelegraphOptions.AllyPCBene     ]  = {ConsoleVariable = "spell.allyPlayerBeneficialTelegraphDisplay",    ControlName = "CheckTelegraphsAllyPCBene"},
    [TelegraphsPattern.TelegraphOptions.AllyPCHarm     ]  = {ConsoleVariable = "spell.allyPlayerDetrimentalTelegraphDisplay",   ControlName = "CheckTelegraphsAllyPCHarm"},
}
function TelegraphsPattern:GetSetting(option)
    return Apollo.GetConsoleVariable(self.ControlMapping[option].ConsoleVariable)
end
function TelegraphsPattern:SetSetting(option, value)
    Apollo.SetConsoleVariable(self.ControlMapping[option].ConsoleVariable, value)
end
function TelegraphsPattern:GetCheckValue(option)
    return self.wndEditor:FindChild(self.ControlMapping[option].ControlName):IsChecked()
end
function TelegraphsPattern:SetCheckValue(option, value)
    return self.wndEditor:FindChild(self.ControlMapping[option].ControlName):SetCheck(value)
end
function TelegraphsPattern:Init()
    self.ShowSelf           = self:GetSetting(TelegraphsPattern.TelegraphOptions.SelfTelegraphs)
    self.ShowEnemy          = self:GetSetting(TelegraphsPattern.TelegraphOptions.EnemyTelegraphs)
    self.ShowEnemyNPCBene   = self:GetSetting(TelegraphsPattern.TelegraphOptions.EnemyNPCBene)
    self.ShowEnemyNPCHarm   = self:GetSetting(TelegraphsPattern.TelegraphOptions.EnemyNPCHarm)
    self.ShowEnemyPCBene    = self:GetSetting(TelegraphsPattern.TelegraphOptions.EnemyPCBene)
    self.ShowEnemyPCHarm    = self:GetSetting(TelegraphsPattern.TelegraphOptions.EnemyPCHarm)
    self.ShowAlly           = self:GetSetting(TelegraphsPattern.TelegraphOptions.AllyTelegraphs)
    self.ShowAllyPartyOnly  = self:GetSetting(TelegraphsPattern.TelegraphOptions.AllyPartyOnly)
    self.ShowAllyNPCBene    = self:GetSetting(TelegraphsPattern.TelegraphOptions.AllyNPCBene)
    self.ShowAllyNPCHarm    = self:GetSetting(TelegraphsPattern.TelegraphOptions.AllyNPCHarm)
    self.ShowAllyPCBene     = self:GetSetting(TelegraphsPattern.TelegraphOptions.AllyPCBene)
    self.ShowAllyPCHarm     = self:GetSetting(TelegraphsPattern.TelegraphOptions.AllyPCHarm)
end;
function TelegraphsPattern:InitFromData(data)
    self.ShowSelf           = data.ShowSelf
    self.ShowEnemy          = data.ShowEnemy
    self.ShowEnemyNPCBene   = data.ShowEnemyNPCBene
    self.ShowEnemyNPCHarm   = data.ShowEnemyNPCHarm
    self.ShowEnemyPCBene    = data.ShowEnemyPCBene
    self.ShowEnemyPCHarm    = data.ShowEnemyPCHarm
    self.ShowAlly           = data.ShowAlly
    self.ShowAllyPartyOnly  = data.ShowAllyPartyOnly
    self.ShowAllyNPCBene    = data.ShowAllyNPCBene
    self.ShowAllyNPCHarm    = data.ShowAllyNPCHarm
    self.ShowAllyPCBene     = data.ShowAllyPCBene
    self.ShowAllyPCHarm     = data.ShowAllyPCHarm
end;
function TelegraphsPattern:ToData()
    return {ArticleType        = self.ArticleType,
            ArticleVersion     = 1,
            ShowSelf           = self.ShowSelf,
            ShowEnemy          = self.ShowEnemy,
            ShowEnemyNPCBene   = self.ShowEnemyNPCBene,
            ShowEnemyNPCHarm   = self.ShowEnemyNPCHarm,
            ShowEnemyPCBene    = self.ShowEnemyPCBene,
            ShowEnemyPCHarm    = self.ShowEnemyPCHarm,
            ShowAlly           = self.ShowAlly,
            ShowAllyPartyOnly  = self.ShowAllyPartyOnly,
            ShowAllyNPCBene    = self.ShowAllyNPCBene,
            ShowAllyNPCHarm    = self.ShowAllyNPCHarm,
            ShowAllyPCBene     = self.ShowAllyPCBene,
            ShowAllyPCHarm     = self.ShowAllyPCHarm,
           }
end;
function TelegraphsPattern:Describe()            return "Sets telegraph display options"; end;
function TelegraphsPattern:Apply()
    self:SetSetting(TelegraphsPattern.TelegraphOptions.SelfTelegraphs,  self.ShowSelf         )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.EnemyTelegraphs, self.ShowEnemy        )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.EnemyNPCBene,    self.ShowEnemyNPCBene )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.EnemyNPCHarm,    self.ShowEnemyNPCHarm )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.EnemyPCBene,     self.ShowEnemyPCBene  )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.EnemyPCHarm,     self.ShowEnemyPCHarm  )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.AllyTelegraphs,  self.ShowAlly         )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.AllyPartyOnly,   self.ShowAllyPartyOnly)
    self:SetSetting(TelegraphsPattern.TelegraphOptions.AllyNPCBene,     self.ShowAllyNPCBene  )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.AllyNPCHarm,     self.ShowAllyNPCHarm  )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.AllyPCBene,      self.ShowAllyPCBene   )
    self:SetSetting(TelegraphsPattern.TelegraphOptions.AllyPCHarm,      self.ShowAllyPCHarm   )
end;
function TelegraphsPattern:MakeEditor(wndParent)
    self.wndEditor =  Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorTelegraphs", wndParent, self)
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.SelfTelegraphs,  self.ShowSelf         )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyTelegraphs, self.ShowEnemy        )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyNPCBene,    self.ShowEnemyNPCBene )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyNPCHarm,    self.ShowEnemyNPCHarm )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyPCBene,     self.ShowEnemyPCBene  )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyPCHarm,     self.ShowEnemyPCHarm  )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.AllyTelegraphs,  self.ShowAlly         )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.AllyPartyOnly,   self.ShowAllyPartyOnly)
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.AllyNPCBene,     self.ShowAllyNPCBene  )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.AllyNPCHarm,     self.ShowAllyNPCHarm  )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.AllyPCBene,      self.ShowAllyPCBene   )
    self:SetCheckValue(TelegraphsPattern.TelegraphOptions.AllyPCHarm,      self.ShowAllyPCHarm   )

end;

function TelegraphsPattern:OnCheckChanged(wndHandler, wndControl)
    self.ShowSelf         = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.SelfTelegraphs )
    self.ShowEnemy        = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyTelegraphs)
    self.ShowEnemyNPCBene = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyNPCBene   )
    self.ShowEnemyNPCHarm = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyNPCHarm   )
    self.ShowEnemyPCBene  = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyPCBene    )
    self.ShowEnemyPCHarm  = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.EnemyPCHarm    )
    self.ShowAlly         = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.AllyTelegraphs )
    self.ShowAllyPartyOnly= self:GetCheckValue(TelegraphsPattern.TelegraphOptions.AllyPartyOnly  )
    self.ShowAllyNPCBene  = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.AllyNPCBene    )
    self.ShowAllyNPCHarm  = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.AllyNPCHarm    )
    self.ShowAllyPCBene   = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.AllyPCBene     )
    self.ShowAllyPCHarm   = self:GetCheckValue(TelegraphsPattern.TelegraphOptions.AllyPCHarm     )
end


Articles[Patterns.eArticleType.Telegraphs] = TelegraphsPattern
--endregion Telegraphs

--region Mount
--[[=============================================================================================]]
--[[
local MountPattern = NewArticlePattern(Patterns.eArticleType.Mount)
Articles[Patterns.eArticleType.Mount] = MountPattern
--]]
--endregion Mount

--region Title
--[[=============================================================================================]]
--[[
    Data:
        TitleName   String. Use nil for no title
--]]
local TitlePattern = {}
TitlePattern = NewArticlePattern(Patterns.eArticleType.Title)

function TitlePattern:Init()
    self.TitleName = nil
    for k,v in pairs(CharacterTitle.GetAvailableTitles()) do
        if CharacterTitle.IsActiveTitle(v) == true then self.TitleName = v:GetTitle(); end
    end
end

function TitlePattern:InitFromData(data)
    self.TitleName = data.TitleName
end

function TitlePattern:ToData()
    return {ArticleType = Patterns.eArticleType.Title, ArticleVersion = 1, TitleName = self.TitleName}
end

function TitlePattern:Describe()
    if self.TitleName == nil then return "Turn off character titles"; end;
    return "Use the title '"..self.TitleName.."'"
end

function TitlePattern:Apply()
    if self.TitleName == nil then CharacterTitle.SetTitle(nil); return; end
    for k,v in pairs(CharacterTitle.GetAvailableTitles()) do
        if v:GetTitle() == self.TitleName then CharacterTitle.SetTitle(v); break; end
    end
end

function TitlePattern:MakeEditor(wndParent)
    local tOptions = {}
    for i,v in ipairs(CharacterTitle.GetAvailableTitles()) do
        tOptions[i] = {Text = v:GetTitle(), Data = v:GetTitle()}
    end
    table.sort(tOptions, function(lhs, rhs) return lhs.Text < rhs.Text; end)
    table.insert(tOptions, 1, {Text = "(no title)", Data = nil})
    ArticleHelpers:MakeSelectEditor(wndParent, "Use title:", tOptions, self.TitleName, self)
end

function TitlePattern:OnXCSelect(control_data, row_data)
    self.TitleName = row_data.Data
end

Articles[Patterns.eArticleType.Title] = TitlePattern
--endregion Title

--region Holomark
--[[=============================================================================================]]
--[[
    Data:
        HoloLeft    Boolean
        HoloBack    Boolean
        HoloRight   Boolean
        HoloNear    Boolean
--]]
local GuildHoloPattern = {}
GuildHoloPattern = NewArticlePattern(Patterns.eArticleType.GuildHolo)

function GuildHoloPattern:Init()
    self.HoloLeft  = GameLib.GetGuildHolomarkVisible(GameLib.GuildHolomark.Left)
    self.HoloRight = GameLib.GetGuildHolomarkVisible(GameLib.GuildHolomark.Right)
    self.HoloBack  = GameLib.GetGuildHolomarkVisible(GameLib.GuildHolomark.Back)
    self.HoloNear  = GameLib.GetGuildHolomarkDistance() -- Returns true for "near"
end

function GuildHoloPattern:InitFromData(data)
    self.HoloLeft  = data.HoloLeft
    self.HoloRight = data.HoloRight
    self.HoloBack  = data.HoloBack
    self.HoloNear  = data.HoloNear
end

function GuildHoloPattern:ToData()
    return  {
                ArticleType    = self.ArticleType,
                ArticleVersion = 1,
                HoloNear       = self.HoloNear,
                HoloLeft       = self.HoloLeft,
                HoloBack       = self.HoloBack,
                HoloRight      = self.HoloRight
            }
end

function GuildHoloPattern:Describe()
    local display = 0   -- Left, Right, Back
    if self.HoloLeft  == true then display = display + 4; end
    if self.HoloRight == true then display = display + 2; end
    if self.HoloBack  == true then display = display + 1; end

    local rettext =
    {
        --[[LRB]]
        --[[000]]   [0] = "Hide all Guild Holomarks",
        --[[001]]   [1] = "Show the Back Guild Holomark",
        --[[010]]   [2] = "Show the Right Guild Holomark",
        --[[011]]   [3] = "Show the Right and Back Guild Holomarks",
        --[[100]]   [4] = "Show the Left Guild Holomark",
        --[[101]]   [5] = "Show the Left and Back Guild Holomarks",
        --[[110]]   [6] = "Show the Left and Right Guild Holomarks",
        --[[111]]   [7] = "Show the Left, Right, and Back Guild Holomarks"
    }

    local ret = rettext[display]

    if display > 0 then
        local disttext =
        {
            [true]  = "(near)",
            [false] = "(far)"
        }
        ret = ret.." "..disttext[self.HoloNear]
    end

    return ret
end

function GuildHoloPattern:Apply()
    GameLib.ShowGuildHolomark(self.HoloLeft, self.HoloRight, self.HoloBack, self.HoloNear)
end

function GuildHoloPattern:MakeEditor(wndParent)
    self.wndEditor =  Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorGuildHolomark", wndParent, self)
    self.wndEditor:FindChild("CheckHoloRight"):SetCheck(self.HoloRight)
    self.wndEditor:FindChild("CheckHoloLeft"):SetCheck(self.HoloLeft)
    self.wndEditor:FindChild("CheckHoloBack"):SetCheck(self.HoloBack)
    self.wndEditor:FindChild("CheckHoloNear"):SetCheck(self.HoloNear)
end

function GuildHoloPattern:ReadConfig()
    self.HoloRight = self.wndEditor:FindChild("CheckHoloRight"):IsChecked()
    self.HoloLeft = self.wndEditor:FindChild("CheckHoloLeft"):IsChecked()
    self.HoloBack = self.wndEditor:FindChild("CheckHoloBack"):IsChecked()
    self.HoloNear = self.wndEditor:FindChild("CheckHoloNear"):IsChecked()
end

function GuildHoloPattern:OnHoloLeftCheck   (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloLeftUncheck (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloBackCheck   (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloBackUncheck (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloRightCheck  (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloRightUncheck(wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloNearCheck   (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end
function GuildHoloPattern:OnHoloNearUncheck (wndHandler, wndControl, eMouseButton) self:ReadConfig(); end

Articles[Patterns.eArticleType.GuildHolo] = GuildHoloPattern
--endregion Holomark

--region PvP
--[[=============================================================================================]]
--[[
    Data:
        FlagState   Boolean
--]]
local PvPPattern = {}
PvPPattern = NewArticlePattern(Patterns.eArticleType.PvP)

function PvPPattern:Init()
    self.FlagState = GameLib.IsPvpFlagged()
end;

function PvPPattern:InitFromData(data)
    self.FlagState = data.FlagState
end;

function PvPPattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, FlagState = self.FlagState}
end;

function PvPPattern:Describe()
    local rettable =
    {
        [true] = "Turn PvP flag on",
        [false] = "Turn PvP flag off",
    }
    return rettable[self.FlagState]
end;

function PvPPattern:Apply()
    local tFlagInfo = GameLib.GetPvpFlagInfo()
    if self.FlagState ~= tFlagInfo.bIsFlagged then
        GameLib.TogglePvpFlags()
--[[
        -- Now tell the user if we couldn't unflag for some reason
        if self.FlagState == false and tFlagInfo.nCooldown > 0 then
            -- They have already tried to unflag already, so a cooldown is progressing
        end
--]]
    end
end

function PvPPattern:MakeEditor(wndParent)
    ArticleHelpers:MakeSelectEditor(wndParent, "Enable or disable PvP", {{Text="Enable (flag for PvP)", Data=true}, {Text="Disable (unflag for PvP if possible)", Data=false}}, self.FlagState, self)
end

function PvPPattern:OnXCSelect(control_data, row_data)
    self.FlagState = row_data.Data
end

Articles[Patterns.eArticleType.PvP] = PvPPattern
--endregion PvP

--region Affiliation
--[[=============================================================================================]]
--[[
    Data
        AffiliationName String
--]]
local AffiliationPattern = {}
AffiliationPattern = NewArticlePattern(Patterns.eArticleType.Affiliation)

function AffiliationPattern:Init()
    self.AffiliationName = nil
    for k,v in pairs(GuildLib.GetGuilds()) do
        if v:IsNameplate() == true then self.AffiliationName = v:GetName(); break; end
    end
end

function AffiliationPattern:InitFromData(data)
    self.AffiliationName = data.AffiliationName
end

function AffiliationPattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, AffiliationName = self.AffiliationName}
end

function AffiliationPattern:Apply()
    for k,v in pairs(GuildLib.GetGuilds()) do
        if v:GetName() == self.AffiliationName then v:SetAsNameplate(); break; end
    end

end

function AffiliationPattern:Describe()
    if self.AffiliationName == nil then return "(error)"; end
    return "Set nameplate affiliation to '"..self.AffiliationName.."'"
end

function AffiliationPattern:MakeEditor(wndParent)
    local tOptions = {}
    for i,v in ipairs(GuildLib.GetGuilds()) do
        tOptions[i] = {Text = v:GetName(), Data = v:GetName()}
    end
    ArticleHelpers:MakeSelectEditor(wndParent, "Affiliation: ", tOptions, self.AffiliationName, self)
end

function AffiliationPattern:OnXCSelect(control_data, row_data)
    self.AffiliationName = row_data.Data
end

Articles[Patterns.eArticleType.Affiliation] = AffiliationPattern
--endregion Affiliation

--region Nameplates
--[[=============================================================================================]]
--[[
    Okay. This one is a bit of a doozey. Loads of options, all pulled from the member variables of
    another addon, and some weird combinations that need special handling.

    Data:
        ShowNameplatesOn
            Self
            Target
            Group
            Guild
            QuestNPCs
            RedNPCs
            YellowNPCs
            GreenNPCs
            FriendlyPCs
            Taxis
            Vendors
        NameplateOptions
            HighlightTarget
            UseOcclusion
            HideInCombat
        NameplateDetails
            NameTarget
            NameNonTarget
            RewardsTarget
            RewardsNonTarget
            GuildNameTarget
            Title
            HealthBarTarget
            HealthBarNonTarget      {Always, Damaged, Never}
        NameplateDecorations
            CastBarTarget
            CastBarNonTarget
            Friend
            Rival
            CertainDeath
            QuestTarget
            MissionTarget
            ChallengeTarget
            PublicEventTarget
--]]
local NameplatesPattern = {}
NameplatesPattern = NewArticlePattern(Patterns.eArticleType.Nameplates)

NameplatesPattern.eShowNonTargetHealthBar = {Never = 1, WhenDamaged = 2, Always = 3} -- Would like Never to be 0, but let's not confuse poor Lua
NameplatesPattern.ShowNTHBComboValues =
{
    [NameplatesPattern.eShowNonTargetHealthBar.Never      ] = {Text="Never",        Data=NameplatesPattern.eShowNonTargetHealthBar.Never      },
    [NameplatesPattern.eShowNonTargetHealthBar.WhenDamaged] = {Text="When Damaged", Data=NameplatesPattern.eShowNonTargetHealthBar.WhenDamaged},
    [NameplatesPattern.eShowNonTargetHealthBar.Always     ] = {Text="Always",       Data=NameplatesPattern.eShowNonTargetHealthBar.Always     },
}
-- Normally I'd avoid these string based mappings as being hard to pick up quickly when doing maintenance
-- six months down the line, but there are a *lot* of these and it makes for fewer bugs.
NameplatesPattern.Mappings =
{
        ShowNameplatesOn =
        {
            Self                = {LocalProp="Self"       , NameplatesProp="bShowMyNameplate"              , ControlName = "CheckNameplateMine"},
            Group               = {LocalProp="Group"      , NameplatesProp="bShowMainGroupOnly"            , ControlName = "CheckNameplateGroup"},
            Guild               = {LocalProp="Guild"      , NameplatesProp="bShowOrganization"             , ControlName = "CheckNameplateOrg"},
            QuestNPCs           = {LocalProp="QuestNPCs"  , NameplatesProp="bShowMainObjectiveOnly"        , ControlName = "CheckNameplateQuest"},
            RedNPCs             = {LocalProp="RedNPCs"    , NameplatesProp="bShowDispositionHostile"       , ControlName = "CheckNameplateHostile"},
            YellowNPCs          = {LocalProp="YellowNPCs" , NameplatesProp="bShowDispositionNeutral"       , ControlName = "CheckNameplateNeutral"},
            GreenNPCs           = {LocalProp="GreenNPCs"  , NameplatesProp="bShowDispositionFriendly"      , ControlName = "CheckNameplateFriendly"},
            FriendlyPCs         = {LocalProp="FriendlyPCs", NameplatesProp="bShowDispositionFriendlyPlayer", ControlName = "CheckNameplateFriendlyPCs"},
            Taxis               = {LocalProp="Taxis"      , NameplatesProp="bShowTaxi"                     , ControlName = "CheckNameplateTaxis"},
            Vendors             = {LocalProp="Vendors"    , NameplatesProp="bShowVendor"                   , ControlName = "CheckNameplateVendors"},
        },
        NameplateOptions =
        {
            HighlightTarget     = {LocalProp="HighlightTarget"  , NameplatesProp="bShowMarkerTarget"         , ControlName = "CheckHighlightTarget"},
            UseOcclusion        = {LocalProp="UseOcclusion"     , NameplatesProp="bUseOcclusion"             , ControlName = "CheckNameplateHideInCombat"},
            HideInCombat        = {LocalProp="HideInCombat"     , NameplatesProp="bHideInCombat"             , ControlName = "CheckNameplateOcclude"},
        },
        NameplateDetails =
        {
            NameTarget          = {LocalProp="NameTarget"      , NameplatesProp="bShowNameTarget"     , ControlName = "CheckNameTarget"},
            NameNonTarget       = {LocalProp="NameNonTarget"   , NameplatesProp="bShowNameMain"       , ControlName = "CheckNameNonTarget"},
            RewardsTarget       = {LocalProp="RewardsTarget"   , NameplatesProp="bShowRewardsTarget"  , ControlName = "CheckRewardIconsTarget"},
            RewardsNonTarget    = {LocalProp="RewardsNonTarget", NameplatesProp="bShowRewardsMain"    , ControlName = "CheckRewardIconsNonTarget"},
            GuildNameTarget     = {LocalProp="GuildNameTarget" , NameplatesProp="bShowGuildNameTarget", ControlName = "CheckGuildNameTarget"},
            Title               = {LocalProp="Title"           , NameplatesProp="bShowTitle"          , ControlName = "CheckShowTitle"},
            HealthBarTarget     = {LocalProp="HealthBarTarget" , NameplatesProp="bShowHealthTarget"   , ControlName = "CheckHealthBarTarget"},
--            HealthBarNonTarget      {Always, Damaged, Never}  -- Not handled by checkbox-mapping
        },
        NameplateDecorations =
        {
            CastBarTarget       = {LocalProp="CastBarTarget"    , NameplatesProp="bShowCastBarTarget"        , ControlName = "CheckCastBarTarget"},
            CastBarNonTarget    = {LocalProp="CastBarNonTarget" , NameplatesProp="bShowCastBarMain"          , ControlName = "CheckCastBarNonTarget"},
            Friend              = {LocalProp="Friend"           , NameplatesProp="bShowFriends"              , ControlName = "CheckShowFriend"},
            Rival               = {LocalProp="Rival"            , NameplatesProp="bShowRivals"               , ControlName = "CheckShowRival"},
            CertainDeath        = {LocalProp="CertainDeath"     , NameplatesProp="bShowCertainDeathMain"     , ControlName = "CheckShowCertainDeath"},
            QuestTarget         = {LocalProp="QuestTarget"      , NameplatesProp="bShowRewardTypeQuest"      , ControlName = "CheckShowQuestTarget"},
            MissionTarget       = {LocalProp="MissionTarget"    , NameplatesProp="bShowRewardTypeMission"    , ControlName = "CheckShowMissionTarget"},
            ChallengeTarget     = {LocalProp="ChallengeTarget"  , NameplatesProp="bShowRewardTypeChallenge"  , ControlName = "CheckShowChallengeTarget"},
            PublicEventTarget   = {LocalProp="PublicEventTarget", NameplatesProp="bShowRewardTypePublicEvent", ControlName = "CheckShowPublicEvent"},
        },
}
function NameplatesPattern:Init()
    local n = Apollo.GetAddon("Nameplates")
    if n == nil then Print("SuitUp: Error reaching the Carbine Nameplates addon. Nameplates actions in suits cannot be applied."); return; end;

    self.ShowNameplatesOn     = {}
    self.NameplateOptions     = {}
    self.NameplateDetails     = {}
    self.NameplateDecorations = {}

    local fnGetNameplateProps = function(tMappings, tSource, tDestination)
        for k,v in pairs(tMappings) do
            tDestination[v.LocalProp] = n[v.NameplatesProp]
            if tDestination[v.LocalProp] == nil then Print("SuitUp: Error: Nameplates property '"..tostring(v.NameplatesProp).."' was nil"); end
        end
    end

    fnGetNameplateProps(NameplatesPattern.Mappings.ShowNameplatesOn,     n, self.ShowNameplatesOn)
    fnGetNameplateProps(NameplatesPattern.Mappings.NameplateOptions,     n, self.NameplateOptions)
    fnGetNameplateProps(NameplatesPattern.Mappings.NameplateDetails,     n, self.NameplateDetails)
    fnGetNameplateProps(NameplatesPattern.Mappings.NameplateDecorations, n, self.NameplateDecorations)

    -- Set self.NameplateDetails.NonTargetHealthBar to one of the enumeration based on props in n
    -- The logic is:
    --  1) If bShowHealthMain is true, use "Always"
    --  2) If bShowHealthMainDamaged is true, use "When Damaged"
    --  3) Use "Never"
   self.NameplateDetails.NonTargetHealthBar = n.bShowHealthMain        and NameplatesPattern.eShowNonTargetHealthBar.Always      or
                                              n.bShowHealthMainDamaged and NameplatesPattern.eShowNonTargetHealthBar.WhenDamaged or
                                                                           NameplatesPattern.eShowNonTargetHealthBar.Never

end;

function NameplatesPattern:InitFromData(data)
    --TODO: Actually do this member by member for version handing
    --TODO: Check version
    self.ShowNameplatesOn = data.ShowNameplatesOn
    self.NameplateOptions = data.NameplateOptions
    self.NameplatesDetails = data.NameplateDetails
    self.NameplateDecorations = data.NameplateDecorations
end;

function NameplatesPattern:ToData()
    return {
            ArticleType          = self.ArticleType,
            ArticleVersion       = 1,
            ShowNameplatesOn     = self.ShowNameplatesOn,
            NameplateOptions     = self.NameplateOptions,
            NameplatesDetails    = self.NameplateDetails,
            NameplateDecorations = self.NameplateDecorations
           }
end;

function NameplatesPattern:Describe()            return "Sets Nameplate display options"; end;

function NameplatesPattern:IsEnabled()
    -- Only enable if the default Nameplates addon is around. If they have a replacement nameplates manager, this Article is useless.
    return nil ~= Apollo.GetAddon("Nameplates")
end

function NameplatesPattern:Apply()
    -- If we are disabled (probably because the Nameplates addon has been replaced), don't try to Apply (but don't delete settings)
    if not self:IsEnabled() then 
        Print("SuitUp: Nameplate Display Action aborted because default Carbine Nameplates not found.")
        return; 
    end

    local n = Apollo.GetAddon("Nameplates")
    if n == nil then Print("SuitUp: Error reaching the Carbin Nameplates addon. Nameplates actions in suits cannot be applied."); return; end;

    local fnAssignBools = function(tMapping, tSource)
        for k,v in pairs(tMapping) do
            if type(tSource[v.LocalProp]) == TYPE.BOOL then
                n[v.NameplatesProp] = tSource[v.LocalProp]
            end
        end
    end

    fnAssignBools(NameplatesPattern.Mappings.ShowNameplatesOn,     self.ShowNameplatesOn)
    fnAssignBools(NameplatesPattern.Mappings.NameplateOptions,     self.NameplateOptions)
    fnAssignBools(NameplatesPattern.Mappings.NameplateDetails,     self.NameplateDetails)
    fnAssignBools(NameplatesPattern.Mappings.NameplateDecorations, self.NameplateDecorations)

    --Set props in n based on the value of self.NameplateDetails.NonTargetHealthBar. The parens here aren't necessary, but they help me read it
   n.bShowHealthMain        = (self.NameplateDetails.NonTargetHealthBar == NameplatesPattern.eShowNonTargetHealthBar.Always)
   n.bShowHealthMainDamaged = (self.NameplateDetails.NonTargetHealthBar == NameplatesPattern.eShowNonTargetHealthBar.Always)      and false or
                              (self.NameplateDetails.NonTargetHealthBar == NameplatesPattern.eShowNonTargetHealthBar.WhenDamaged) and true or
                                                                                                                                      false
end;

function NameplatesPattern:MakeEditor(wndParent)
    if not self:IsEnabled() then 
        Print("SuitUp: Cannot configure Nameplates action because the Carbine Nameplates addon cannot be found. Are you running a replacement nameplate addon?")
        return; 
    end;
    self.wndEditor =  Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorNameplates", wndParent, self)

    local fnSetChecks = function(tMapping, tSource)
        for k,v in pairs(tMapping) do
            local w = self.wndEditor:FindChild(v.ControlName)
            if w == nil then Print("SuitUp: Error: Control '"..tostring(v.ControlName).."' not found for property ..'"..tostring(v.LocalProp).."'"); break; end

            w:SetCheck(tSource[v.LocalProp])
        end
    end

    fnSetChecks(NameplatesPattern.Mappings.ShowNameplatesOn,     self.ShowNameplatesOn)
    fnSetChecks(NameplatesPattern.Mappings.NameplateOptions,     self.NameplateOptions)
    fnSetChecks(NameplatesPattern.Mappings.NameplateDetails,     self.NameplateDetails)
    fnSetChecks(NameplatesPattern.Mappings.NameplateDecorations, self.NameplateDecorations)

    -- Set up the NTHB button
    local wndNTHB = self.wndEditor:FindChild("ButtonNonTargetHealthBar")
    wndNTHB:SetText(NameplatesPattern.ShowNTHBComboValues[self.NameplateDetails.NonTargetHealthBar].Text)
    wndNTHB:SetData(self.NameplateDetails.NonTargetHealthBar)
end;

function NameplatesPattern:OnCheckChanged(wndHandler, wndControl)
    local fnGetChecks = function(tMapping, tDestination)
        for k,v in pairs(tMapping) do
            local w = self.wndEditor:FindChild(v.ControlName)
            if w == nil then Print("SuitUp: Error: Control '"..tostring(v.ControlName).."' not found for property ..'"..tostring(v.LocalProp).."'"); break; end

            tDestination[v.LocalProp] = w:IsChecked()
        end
    end

    fnGetChecks(NameplatesPattern.Mappings.ShowNameplatesOn,     self.ShowNameplatesOn)
    fnGetChecks(NameplatesPattern.Mappings.NameplateOptions,     self.NameplateOptions)
    fnGetChecks(NameplatesPattern.Mappings.NameplateDetails,     self.NameplateDetails)
    fnGetChecks(NameplatesPattern.Mappings.NameplateDecorations, self.NameplateDecorations)
end

function NameplatesPattern:OnNTHBChange(wndHandler, wndControl, eMouseButton)
    -- Hardcoding 3 for the number of states because it's a PITA to get a count
    self.NameplateDetails.NonTargetHealthBar = self.NameplateDetails.NonTargetHealthBar % 3 + 1

    local wndNTHB = self.wndEditor:FindChild("ButtonNonTargetHealthBar")
    wndNTHB:SetText(NameplatesPattern.ShowNTHBComboValues[self.NameplateDetails.NonTargetHealthBar].Text)
    wndNTHB:SetData(self.NameplateDetails.NonTargetHealthBar)
end

Articles[Patterns.eArticleType.Nameplates] = NameplatesPattern
--endregion Nameplates

--region Macro
--[[=============================================================================================]]
--[[
    Data:
        MacroId     Number
--]]
local MacroPattern = {}
MacroPattern = NewArticlePattern(Patterns.eArticleType.Macro)

function MacroPattern:Init()
    self.MacroId = nil
end;

function MacroPattern:InitFromData(data)
    self.MacroId = data.MacroId
end;

function MacroPattern:ToData()
    return {ArticleType = self.ArticleType, ArticleVersion = 1, MacroId = self.MacroId}
end;

function MacroPattern:Describe()
    if self.MacroId == nil then return "No macro configured to run"; end

    local macros = MacrosLib.GetMacrosList()
    for i,v in ipairs(macros) do
        if v.nId == self.MacroId then return "Run the macro '"..tostring(v.strName).."'"; end
    end
    return "Unknown macro configured; it will not be run";
end;

function MacroPattern:Apply()
    if self.MacroId ~= nil then MacrosLib.DoMacro(self.MacroId); end
end;

function MacroPattern:MakeEditor(wndParent)
    local tOptions = {}
    local macros = MacrosLib.GetMacrosList()
    for i,v in ipairs(macros) do
        table.insert(tOptions, {Text = v.strName, Data = v.nId})
    end
    ArticleHelpers:MakeSelectEditor(wndParent, "Run which macro:", tOptions, self.MacroId, self)
end;

function MacroPattern:OnXCSelect(control_data, row_data)
    self.MacroId = row_data.Data
end

Articles[Patterns.eArticleType.Macro] = MacroPattern
--endregion Macro

--region Targeting
--[[=============================================================================================]]
--[[
    Data:
        AutoTarget                  spell.disableAutoTargeting          -- Note that this is reversed!
        AutoAimAtTarget             Player.ignoreAlwaysFaceTarget       -- Note that this is probably reversed!
            LockToAutoAimedTarget   Player.disableFacingLock            -- Note that this is reversed!          -- Requires AutoAimAtTarget
        SelfCast                    spell.autoSelectCharacter
        MoveToCast                  Player.moveToTargetOnSelfAOE
        HoldToCharge                spell.useButtonDownForAbilities
            HoldRepeatsCasting      spell.holdToContinueCasting         --(Requires HoldToCharge)
        StickyTargeting             player.stickyTargeting
--]]
local TargetingPattern = {}
TargetingPattern = NewArticlePattern(Patterns.eArticleType.Targeting)

function TargetingPattern:Init()
     self.AutoTarget            = not Apollo.GetConsoleVariable("spell.disableAutoTargeting")
     self.AutoAimAtTarget       = not Apollo.GetConsoleVariable("Player.ignoreAlwaysFaceTarget")
     self.LockToAutoAimedTarget = not Apollo.GetConsoleVariable("Player.disableFacingLock")
     self.SelfCast              =     Apollo.GetConsoleVariable("spell.autoSelectCharacter")
     self.MoveToCast            =     Apollo.GetConsoleVariable("Player.moveToTargetOnSelfAOE")
     self.HoldToCharge          =     Apollo.GetConsoleVariable("spell.useButtonDownForAbilities")
     self.HoldRepeatsCasting    =     Apollo.GetConsoleVariable("spell.holdToContinueCasting")
     self.StickyTargeting       =     Apollo.GetConsoleVariable("player.stickyTargeting")
end;

function TargetingPattern:InitFromData(data)
    self.AutoTarget            = data.AutoTarget
    self.AutoAimAtTarget       = data.AutoAimAtTarget
    self.LockToAutoAimedTarget = data.LockToAutoAimedTarget
    self.SelfCast              = data.SelfCast
    self.MoveToCast            = data.MoveToCast
    self.HoldToCharge          = data.HoldToCharge
    self.HoldRepeatsCasting    = data.HoldRepeatsCasting
    self.StickyTargeting       = data.StickyTargeting
end;

function TargetingPattern:ToData()
    return {
            ArticleType = self.ArticleType,
            ArticleVersion = 1,
            AutoTarget           =  self.AutoTarget,
            AutoAimAtTarget      =  self.AutoAimAtTarget,
            LockToAutoAimedTarget=  self.LockToAutoAimedTarget,
            SelfCast             =  self.SelfCast,
            MoveToCast           =  self.MoveToCast,
            HoldToCharge         =  self.HoldToCharge,
            HoldRepeatsCasting   =  self.HoldRepeatsCasting,
            StickyTargeting      =  self.StickyTargeting,
            }
end;

function TargetingPattern:Describe() return "Sets various targeting options"; end;

function TargetingPattern:Apply()
    Apollo.SetConsoleVariable("spell.disableAutoTargeting",      not self.AutoTarget           )
    Apollo.SetConsoleVariable("Player.ignoreAlwaysFaceTarget",   not self.AutoAimAtTarget      )
    Apollo.SetConsoleVariable("Player.disableFacingLock",        not self.LockToAutoAimedTarget)
    Apollo.SetConsoleVariable("spell.autoSelectCharacter",           self.SelfCast             )
    Apollo.SetConsoleVariable("Player.moveToTargetOnSelfAOE",        self.MoveToCast           )
    Apollo.SetConsoleVariable("spell.holdToContinueCasting",         self.HoldRepeatsCasting   )
    Apollo.SetConsoleVariable("spell.useButtonDownForAbilities",     self.HoldToCharge         ) --NOTE: Has to be set after HoldRepeatsCasting, or turning off HoldRepeatsCasting will turn this off
    Apollo.SetConsoleVariable("player.stickyTargeting",              self.StickyTargeting      )
end;

function TargetingPattern:MakeEditor(wndParent)
    self.wndEditor =  Apollo.LoadForm(Apollo.GetAddon("SuitUp").xmlDoc, "ArticleEditorTargeting", wndParent, self)
    self.wndEditor:FindChild("CheckAutoTarget"):SetCheck(self.AutoTarget)
    self.wndEditor:FindChild("CheckAutoAim"):SetCheck(self.AutoAimAtTarget)
    self.wndEditor:FindChild("CheckDisableRightButtonAiming"):SetCheck(self.LockToAutoAimedTarget)
    self.wndEditor:FindChild("CheckSelfCast"):SetCheck(self.SelfCast)
    self.wndEditor:FindChild("CheckMoveToCast"):SetCheck(self.MoveToCast)
    self.wndEditor:FindChild("CheckHoldToCharge"):SetCheck(self.HoldToCharge)
    self.wndEditor:FindChild("CheckHoldRepeatsCasting"):SetCheck(self.HoldRepeatsCasting)
    self.wndEditor:FindChild("CheckStickyTargeting"):SetCheck(self.StickyTargeting)
end;

function TargetingPattern:OnCheckChanged(wndHandler, wndControl)
    self.AutoTarget            = self.wndEditor:FindChild("CheckAutoTarget"):IsChecked()
    self.AutoAimAtTarget       = self.wndEditor:FindChild("CheckAutoAim"):IsChecked()
    self.LockToAutoAimedTarget = self.wndEditor:FindChild("CheckDisableRightButtonAiming"):IsChecked()
    self.SelfCast              = self.wndEditor:FindChild("CheckSelfCast"):IsChecked()
    self.MoveToCast            = self.wndEditor:FindChild("CheckMoveToCast"):IsChecked()
    self.HoldToCharge          = self.wndEditor:FindChild("CheckHoldToCharge"):IsChecked()
    self.HoldRepeatsCasting    = self.wndEditor:FindChild("CheckHoldRepeatsCasting"):IsChecked()
    self.StickyTargeting       = self.wndEditor:FindChild("CheckStickyTargeting"):IsChecked()
end

Articles[Patterns.eArticleType.Targeting] = TargetingPattern
--endregion Targeting

--[[=============================================================================================]]
Patterns.ArticlePatterns = Articles
return Patterns